@@ -8,7 +8,7 @@ import numpy as np
 from numpy.testing import assert_allclose, assert_array_equal
 import pytest
 
-from contourpy import FillType, contour_generator
+from contourpy import FillType, contour_generator, max_threads
 from contourpy.util.data import random, simple
 
 from . import util_test
@@ -40,6 +40,21 @@ def test_filled_decreasing_levels(name: str) -> None:
         cont_gen.filled(2.0, 1.0)
 
 
+@pytest.mark.parametrize("name", util_test.all_names())
+def test_filled_nan_levels(name: str) -> None:
+    cont_gen = contour_generator(z=[[0, 1], [2, 3]], name=name, fill_type=FillType.OuterCode)
+
+    msg = r"lower_level and upper_level cannot be NaN"
+    with pytest.raises(ValueError, match=msg):
+        cont_gen.filled(0, np.nan)
+
+    with pytest.raises(ValueError, match=msg):
+        cont_gen.filled(np.nan, 0)
+
+    with pytest.raises(ValueError, match=msg):
+        cont_gen.filled(np.nan, np.nan)
+
+
 @pytest.mark.parametrize("name", util_test.all_names())
 def test_filled_identical_levels(name: str) -> None:
     cont_gen = contour_generator(z=[[0, 1], [2, 3]], name=name, fill_type=FillType.OuterCode)
@@ -48,8 +63,9 @@ def test_filled_identical_levels(name: str) -> None:
 
 
 @pytest.mark.image
+@pytest.mark.parametrize("multi", [False, True])
 @pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
-def test_filled_simple(name: str, fill_type: FillType) -> None:
+def test_filled_simple(name: str, fill_type: FillType, multi: bool) -> None:
     from contourpy.util.mpl_renderer import MplTestRenderer
 
     from .image_comparison import compare_images
@@ -63,16 +79,21 @@ def test_filled_simple(name: str, fill_type: FillType) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    if multi:
+        multi_filled = cont_gen.multi_filled(levels)
+        renderer.multi_filled(multi_filled, fill_type)
+    else:
+        for i in range(len(levels)-1):
+            renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
     image_buffer = renderer.save_to_buffer()
 
-    compare_images(image_buffer, "filled_simple.png", f"{name}_{fill_type}")
+    compare_images(image_buffer, "filled_simple.png", f"{name}_{fill_type}_{multi}")
 
 
 @pytest.mark.image
+@pytest.mark.parametrize("multi", [False, True])
 @pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
-def test_filled_simple_chunk(name: str, fill_type: FillType) -> None:
+def test_filled_simple_chunk(name: str, fill_type: FillType, multi: bool) -> None:
     from contourpy.util.mpl_renderer import MplTestRenderer
 
     from .image_comparison import compare_images
@@ -88,12 +109,15 @@ def test_filled_simple_chunk(name: str, fill_type: FillType) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    if multi:
+        renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
+    else:
+        for i in range(len(levels)-1):
+            renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
-        image_buffer, "filled_simple_chunk.png", f"{name}_{fill_type}", mean_threshold=0.12,
+        image_buffer, "filled_simple_chunk.png", f"{name}_{fill_type}_{multi}", mean_threshold=0.12,
     )
 
 
@@ -117,8 +141,7 @@ def test_filled_simple_chunk_threads(fill_type: FillType, thread_count: int) ->
     assert cont_gen.thread_count == thread_count
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
@@ -142,8 +165,7 @@ def test_filled_simple_no_corner_mask(name: str, fill_type: FillType) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(image_buffer, "filled_simple_no_corner_mask.png", f"{name}_{fill_type}")
@@ -167,8 +189,7 @@ def test_filled_simple_no_corner_mask_chunk(name: str, fill_type: FillType) -> N
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
@@ -198,8 +219,7 @@ def test_filled_simple_no_corner_mask_chunk_threads(fill_type: FillType, thread_
     assert cont_gen.thread_count == thread_count
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
@@ -225,8 +245,7 @@ def test_filled_simple_corner_mask(name: str) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(image_buffer, "filled_simple_corner_mask.png", f"{name}_{fill_type}")
@@ -251,8 +270,7 @@ def test_filled_simple_corner_mask_chunk(name: str) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
@@ -283,8 +301,7 @@ def test_filled_simple_corner_mask_chunk_threads(fill_type: FillType, thread_cou
     assert cont_gen.thread_count == thread_count
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
@@ -310,16 +327,16 @@ def test_filled_simple_quad_as_tri(name: str) -> None:
 
     fill_type = cont_gen.fill_type
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(image_buffer, "filled_simple_quad_as_tri.png", f"{name}")
 
 
 @pytest.mark.image
+@pytest.mark.parametrize("multi", [False, True])
 @pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
-def test_filled_random(name: str, fill_type: FillType) -> None:
+def test_filled_random(name: str, fill_type: FillType, multi: bool) -> None:
     from contourpy.util.mpl_renderer import MplTestRenderer
 
     from .image_comparison import compare_images
@@ -333,16 +350,21 @@ def test_filled_random(name: str, fill_type: FillType) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    if multi:
+        multi_filled = cont_gen.multi_filled(levels)
+        renderer.multi_filled(multi_filled, fill_type)
+    else:
+        for i in range(len(levels)-1):
+            renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
     image_buffer = renderer.save_to_buffer()
 
-    compare_images(image_buffer, "filled_random.png", f"{name}_{fill_type}")
+    compare_images(image_buffer, "filled_random.png", f"{name}_{fill_type}_{multi}")
 
 
 @pytest.mark.image
+@pytest.mark.parametrize("multi", [False, True])
 @pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
-def test_filled_random_chunk(name: str, fill_type: FillType) -> None:
+def test_filled_random_chunk(name: str, fill_type: FillType, multi: bool) -> None:
     from contourpy.util.mpl_renderer import MplTestRenderer
 
     from .image_comparison import compare_images
@@ -358,8 +380,11 @@ def test_filled_random_chunk(name: str, fill_type: FillType) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    if multi:
+        renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
+    else:
+        for i in range(len(levels)-1):
+            renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
     image_buffer = renderer.save_to_buffer()
 
     max_threshold = None
@@ -376,7 +401,7 @@ def test_filled_random_chunk(name: str, fill_type: FillType) -> None:
             mean_threshold = 0.19
 
     compare_images(
-        image_buffer, "filled_random_chunk.png", f"{name}_{fill_type}",
+        image_buffer, "filled_random_chunk.png", f"{name}_{fill_type}_{multi}",
         max_threshold=max_threshold, mean_threshold=mean_threshold,
     )
 
@@ -401,8 +426,7 @@ def test_filled_random_chunk_threads(fill_type: FillType, thread_count: int) ->
     assert cont_gen.thread_count == thread_count
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     if fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
@@ -434,8 +458,7 @@ def test_filled_random_no_corner_mask(name: str, fill_type: FillType) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(image_buffer, "filled_random_no_corner_mask.png", f"{name}_{fill_type}")
@@ -459,8 +482,7 @@ def test_filled_random_no_corner_mask_chunk(name: str, fill_type: FillType) -> N
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     max_threshold = None
@@ -503,8 +525,7 @@ def test_filled_random_no_corner_mask_chunk_threads(fill_type: FillType, thread_
     assert cont_gen.thread_count == thread_count
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     if fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
@@ -537,8 +558,7 @@ def test_filled_random_corner_mask(name: str) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(image_buffer, "filled_random_corner_mask.png", f"{name}_{fill_type}")
@@ -563,8 +583,7 @@ def test_filled_random_corner_mask_chunk(name: str) -> None:
     assert cont_gen.thread_count == 1
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     max_threshold = None
@@ -601,8 +620,7 @@ def test_filled_random_corner_mask_chunk_threads(fill_type: FillType, thread_cou
     assert cont_gen.thread_count == thread_count
 
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(
@@ -627,8 +645,7 @@ def test_filled_random_quad_as_tri(name: str) -> None:
 
     fill_type = cont_gen.fill_type
     renderer = MplTestRenderer()
-    for i in range(len(levels)-1):
-        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
+    renderer.multi_filled(cont_gen.multi_filled(levels), fill_type)
     image_buffer = renderer.save_to_buffer()
 
     compare_images(image_buffer, "filled_random_quad_as_tri.png", f"{name}")
@@ -734,7 +751,10 @@ def test_return_by_fill_type(
 @pytest.mark.threads
 @pytest.mark.parametrize("fill_type", FillType.__members__.values())
 @pytest.mark.parametrize("name, thread_count",
-                         [("serial", 1), ("threaded", 1), ("threaded", 2)])
+                         [("serial", 1), ("threaded", 1),
+                         pytest.param("threaded", 2,
+                            marks = pytest.mark.skipif(
+                            max_threads() <= 1, reason = "executing on unicore host"))])
 def test_return_by_fill_type_chunk(
     xyz_chunk_test: tuple[cpy.CoordinateArray, ...],
     name: str,
@@ -902,7 +922,7 @@ def test_filled_compare_slow(seed: int) -> None:
         code_list = filled_mpl2014[1]
         n_points = reduce(add, map(len, code_list))
         n_outers = len(code_list)
-        n_lines = reduce(add, map(lambda c: np.count_nonzero(c == 1), code_list))
+        n_lines = reduce(add, (np.count_nonzero(c == 1) for c in code_list))
 
         assert filled_serial[0][0] is not None
         assert n_points == len(filled_serial[0][0])
@@ -916,8 +936,8 @@ def test_filled_compare_slow(seed: int) -> None:
 
 @pytest.mark.parametrize("name", util_test.all_names())
 @pytest.mark.parametrize("z", [np.nan, -np.nan, np.inf, -np.inf])
-@pytest.mark.parametrize("lower_level", [0.0, np.nan, -np.nan, np.inf, -np.inf])
-@pytest.mark.parametrize("upper_level", [0.0, np.nan, -np.nan, np.inf, -np.inf])
+@pytest.mark.parametrize("lower_level", [0.0, np.inf, -np.inf])
+@pytest.mark.parametrize("upper_level", [0.0, np.inf, -np.inf])
 def test_filled_z_nonfinite(name: str, z: float, lower_level: float, upper_level: float) -> None:
     cont_gen = contour_generator(z=[[z, z], [z, z]], name=name, fill_type=FillType.OuterCode)
     if lower_level >= upper_level:
@@ -951,3 +971,73 @@ def test_filled_infinite_level(name: str) -> None:
         points = filled[0][0]
         assert isinstance(points, np.ndarray)
         assert_allclose(points, expected)
+
+
+@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
+def test_multi_filled_invalid_levels(name: str, fill_type: FillType) -> None:
+    z = [[1, 2], [3, 4]]
+    cont_gen = contour_generator(z=z, name=name, fill_type=fill_type)
+
+    with pytest.raises(ValueError, match=r"Levels array must be 1D not 2D"):
+        cont_gen.multi_filled([[1, 2]])
+
+    with pytest.raises(ValueError, match=r"Levels array must be 1D not 3D"):
+        cont_gen.multi_filled([[[1], [2]]])
+
+    with pytest.raises(ValueError, match=r"Levels array must have at least 2 elements, not 1"):
+        cont_gen.multi_filled([1])
+
+    with pytest.raises(ValueError, match=r"Levels array must have at least 2 elements, not 0"):
+        cont_gen.multi_filled([])
+
+    for levels in (
+        [0, np.nan],
+        [np.nan, 0],
+        [np.nan, np.nan],
+    ):
+        with pytest.raises(ValueError, match=r"Levels must not contain any NaN"):
+            cont_gen.multi_filled(levels)
+
+    for levels in (
+        [1.0, 1],
+        [1, 0.99],
+        [0, -np.inf],
+        [np.inf, 0],
+        [np.inf, -np.inf],
+    ):
+        with pytest.raises(ValueError, match=r"Levels must be increasing"):
+            cont_gen.multi_filled(levels)
+
+
+@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
+@pytest.mark.parametrize("corner_mask", [None, False, True])
+def test_multi_filled_levels_type(name: str, fill_type: FillType, corner_mask: bool | None) -> None:
+    if corner_mask and name in ["mpl2005", "mpl2014"]:
+        pytest.skip()
+
+    z = random((10, 10), mask_fraction=0.0 if corner_mask is None else 0.05)[2]*4
+    cont_gen = contour_generator(z=z, name=name, corner_mask=corner_mask, fill_type=fill_type)
+    levels = [2, 3, 4]
+
+    # Reference result using filled() not multi_filled().
+    reference = [cont_gen.filled(lower, upper) for lower, upper in zip(levels[:-1], levels[1:])]
+
+    # List.
+    util_test.assert_equal_recursive(reference, cont_gen.multi_filled(list(levels)))
+
+    # Tuple.
+    util_test.assert_equal_recursive(reference, cont_gen.multi_filled(tuple(levels)))
+
+    # Numpy array of floats.
+    util_test.assert_equal_recursive(
+        reference, cont_gen.multi_filled(np.asarray(levels, dtype=np.float64)),
+    )
+
+    # Numpy array of ints.
+    util_test.assert_equal_recursive(
+        reference, cont_gen.multi_filled(np.asarray(levels, dtype=np.int64)),
+    )
+
+    # Slice of larger numpy array.
+    levels2 = np.array([1.5, 2, 2.5, 3, 3.5, 4, 4.5])
+    util_test.assert_equal_recursive(reference, cont_gen.multi_filled(levels2[1::2]))
